# wave equation with cartesian PML and DG
# CAA2 category1 p2
# sound scattering around a circle
# MODAV PML works very good
# PML model [Modave, Lambrechts, Geuzaine, 2016]
# 
# p_t + div(u) +(s1+s2)p + s1 s2 pbar = 0
# u_t + grad(p) + [(s1-s2)(ux+uxbar); -(s1-s2)(uy+uybar)] = 0
# pbar_t -p = 0
# ubar_t +  [s2 (ux+uxbar); s1(uy+uybar)] = 0

from ngsolve.comp import *
from ngsolve import *
comm = MPI_Init()
rank = comm.rank
ngsglobals.msg_level=0

from netgen.geom2d import SplineGeometry
import netgen.meshing as netgen


###################################################################
# polynomial degree/cfl/mesh size
k = 2
cfl = 0.05

h0 = 0.1
tau = h0*cfl

# domain and time steps
xr = 6    # domain is [-xr, xr]x[-xr, xr]
rad = 0.5 # domain with out a circle with rad = .5

# PML parameters
dd = 6*h0; sigma=10
s1 = CoefficientFunction([0, sigma, 0, sigma])
s2 = CoefficientFunction([0, 0, sigma, sigma])

tend = 10
t = 0

cc = True
count = 0

# source term: p(0) = source
source = exp(-log(2)*((x-4)**2+y**2)/0.04)
###################################################################




## set up domain/mesh
geo = SplineGeometry()
pntx = [-dd-xr,-xr,xr,xr+dd]
pnty = [-dd-xr,-xr,xr,xr+dd]
pmlpnts = []
for yi in pnty:
    for xi in pntx:
        pmlpnts.append (geo.AddPoint(xi,yi))
#inner rect
geo.Append (["line", pmlpnts[5], pmlpnts[6]], leftdomain=1, rightdomain=3)
geo.Append (["line", pmlpnts[6], pmlpnts[10]], leftdomain=1, rightdomain=2)
geo.Append (["line", pmlpnts[10], pmlpnts[9]], leftdomain=1, rightdomain=3)
geo.Append (["line", pmlpnts[9], pmlpnts[5]], leftdomain=1, rightdomain=2)

# right x-pml
geo.Append (["line", pmlpnts[6], pmlpnts[7]], leftdomain=2, rightdomain=4)
geo.Append (["line", pmlpnts[7], pmlpnts[11]], leftdomain=2, rightdomain=0,
    bc="pec")
geo.Append (["line", pmlpnts[11], pmlpnts[10]], leftdomain=2, rightdomain=4)
# bottom y-pml
geo.Append (["line", pmlpnts[5], pmlpnts[1]], leftdomain=3, rightdomain=4)
geo.Append (["line", pmlpnts[1], pmlpnts[2]], leftdomain=3, rightdomain=0,
    bc = "pec")
geo.Append (["line", pmlpnts[2], pmlpnts[6]], leftdomain=3, rightdomain=4)
# top y-pml
geo.Append (["line", pmlpnts[10], pmlpnts[14]], leftdomain=3, rightdomain=4)
geo.Append (["line", pmlpnts[14], pmlpnts[13]], leftdomain=3,
    rightdomain=0, bc = "pec")
geo.Append (["line", pmlpnts[13], pmlpnts[9]], leftdomain=3, rightdomain=4)
# left x-pml
geo.Append (["line", pmlpnts[9], pmlpnts[8]], leftdomain=2, rightdomain=4)
geo.Append (["line", pmlpnts[8], pmlpnts[4]], leftdomain=2, 
    rightdomain=0, bc = "pec")
geo.Append (["line", pmlpnts[4], pmlpnts[5]], leftdomain=2, rightdomain=4)

# bottom right xy-pml
geo.Append (["line", pmlpnts[2], pmlpnts[3]], leftdomain=4, 
    rightdomain=0, bc = "pec")
geo.Append (["line", pmlpnts[3], pmlpnts[7]], leftdomain=4, 
    rightdomain=0, bc = "pec")
# top right xy-pml
geo.Append (["line", pmlpnts[11], pmlpnts[15]], leftdomain=4, 
    rightdomain=0, bc = "pec")
geo.Append (["line", pmlpnts[15], pmlpnts[14]], leftdomain=4, 
    rightdomain=0, bc = "pec")
# top left xy-pml
geo.Append (["line", pmlpnts[13], pmlpnts[12]], leftdomain=4, 
    rightdomain=0, bc = "pec")
geo.Append (["line", pmlpnts[12], pmlpnts[8]], leftdomain=4, 
    rightdomain=0, bc = "pec")
# bottom left xy-pml
geo.Append (["line", pmlpnts[4], pmlpnts[0]], leftdomain=4, 
    rightdomain=0, bc = "pec")
geo.Append (["line", pmlpnts[0], pmlpnts[1]], leftdomain=4, 
    rightdomain=0, bc = "pec")

geo.AddCircle(c=(0,0), r=rad, leftdomain=0, rightdomain=1,bc="pec" )

geo.SetMaterial(1,"inner")
geo.SetMaterial(2,"pmlx")
geo.SetMaterial(3,"pmly")
geo.SetMaterial(4,"pmlxy")

if rank==0:
    ngmesh = geo.GenerateMesh(maxh=h0)
    ngmesh.Distribute(comm)
else:
    ngmesh = netgen.Mesh.Receive(comm)
    ngmesh.SetGeometry(geo)
mesh = Mesh(ngmesh)
mesh.Curve(k)


## FESpaces and bilinear forms
fes_u = VectorL2(mesh, order=k)
fes_p = L2(mesh, order=k) 

fes_u2 = VectorL2(mesh, order=k, definedon=mesh.Materials("pmlx|pmly|pmlxy"))
fes_p2 = L2(mesh, order=k, definedon=mesh.Materials("pmlx|pmly|pmlxy")) 

fes = FESpace( [fes_p,fes_p2,fes_u, fes_u2] )
p,pbar, u,ubar = fes.TrialFunction()
q,qbar, v,vbar = fes.TestFunction()

n = specialcf.normal(2) 
# proxy functions
ux = u[0]
uy = u[1]
vx = v[0]
vy = v[1]
uxbar = ubar[0]
uybar = ubar[1]
vxbar = vbar[0]
vybar = vbar[1]
# numerical fluxes
phat = 0.5*(p+p.Other())+0.5*(u-u.Other())*n
uhatn = 0.5*(u+u.Other())*n +0.5*(p-p.Other())


# bilinear form terms
if cc :
  # volume part
  term1 = (-p*div(v)-grad(q)*u).Compile(True, True) 
  # interior edges
  term2 = (phat *(n*(v-v.Other()))+uhatn*(q-q.Other())).Compile(True, True)
  # bdry edges
  term3 = (p * (n*v)).Compile(True, True)
  # pmlx
  term4 =(s1*p*q+
    s1*(ux+uxbar)*vx-s1*(uy+uybar)*vy-p*qbar
    +s1*(uy+uybar)*vybar).Compile(True, True)
  # pmly
  term5 =(s2*p*q
    -s2*(ux+uxbar)*vx+s2*(uy+uybar)*vy-p*qbar
    +s2*(ux+uxbar)*vxbar).Compile(True, True)
  # pmlxy
  term6 =((s1+s2)*p*q+s1*s2*pbar*q+
    (s1-s2)*(ux+uxbar)*vx-(s1-s2)*(uy+uybar)*vy-p*qbar
    +s2*(ux+uxbar)*vxbar
    +s1*(uy+uybar)*vybar).Compile(True, True)
else :
  # volume part
  term1 = (-p*div(v)-grad(q)*u)
  # interior edges
  term2 = (phat *(n*(v-v.Other()))+uhatn*(q-q.Other()))
  # bdry edges
  term3 = (p * (n*v))
  # pmlx
  term4 =(s1*p*q+
    s1*(ux+uxbar)*vx-s1*(uy+uybar)*vy-p*qbar
    +s1*(uy+uybar)*vybar)
  # pmly
  term5 =(s2*p*q
    -s2*(ux+uxbar)*vx+s2*(uy+uybar)*vy-p*qbar
    +s2*(ux+uxbar)*vxbar)
  # pmlxy
  term6 =((s1+s2)*p*q+s1*s2*pbar*q+
    (s1-s2)*(ux+uxbar)*vx-(s1-s2)*(uy+uybar)*vy-p*qbar
    +s2*(ux+uxbar)*vxbar
    +s1*(uy+uybar)*vybar)

B  = BilinearForm(fes, nonassemble=True)
# regular part
B += SymbolicBFI(term1)
B += SymbolicBFI(term2, VOL, skeleton=True)
B += SymbolicBFI(term3, BND, skeleton=True)

# pml part
B += SymbolicBFI(term4, definedon=mesh.Materials("pmlx"))
B += SymbolicBFI(term5, definedon=mesh.Materials("pmly"))
B += SymbolicBFI(term6, definedon=mesh.Materials("pmlxy"))

gfu = GridFunction(fes)
gfu.components[0].Set (source )

pdofs = fes.Range(0)
pbardofs = fes.Range(1)
udofs = fes.Range(2)
ubardofs = fes.Range(3)

w = gfu.vec.CreateVector()
u0 = gfu.vec.CreateVector()
u1 = gfu.vec.CreateVector()


nsteps = int(tend/tau/100)
if rank == 0:
  print("Total time steps: ", 100*nsteps, flush=True)
  start = comm.WTime()
  start0 = start
with TaskManager():
    while t < tend:
        if rank == 0 and count%nsteps==0:
          end = comm.WTime()
          step = int(count/nsteps)
          print ("percentage:  ", step, "% ",
              "elapsed(s): ", format(end-start,'.3e'),
              " total(m): ", format((end-start0)/60,'.3e'),
              " remain(m): ", format((100-step)*(end-start)/60,'.3e'),
              flush=True)
          start = end
        u0.data = gfu.vec
        # stage 1
        B.Apply (gfu.vec, w)
        fes_p.SolveM (vec=w[pdofs])
        fes_p2.SolveM (vec=w[pbardofs])
        fes_u.SolveM (vec=w[udofs])
        fes_u2.SolveM (vec=w[ubardofs])
        gfu.vec.data -= tau * w

        # stage 2
        B.Apply (gfu.vec, w)
        fes_p.SolveM (vec=w[pdofs])
        fes_p2.SolveM (vec=w[pbardofs])
        fes_u.SolveM (vec=w[udofs])
        fes_u2.SolveM (vec=w[ubardofs])
        u1.data = .75*u0+.25*(gfu.vec-tau*w)
        gfu.vec.data =u1

        # stage 3
        B.Apply (gfu.vec, w)
        fes_p.SolveM (vec=w[pdofs])
        fes_p2.SolveM (vec=w[pbardofs])
        fes_u.SolveM (vec=w[udofs])
        fes_u2.SolveM (vec=w[ubardofs])
        u1.data = 1/3*u0+2/3*(gfu.vec-tau*w)
        gfu.vec.data =u1
        t += tau
        count = count+1;
        comm.Barrier()
